package jwo.landserf.process.proj;

import jwo.landserf.structure.*;         // For coordinate footprints.

//  ***********************************************************
/** Class for representing the Ordnance Survey GB, National Grid
  * coordinate system. This is a UTM transformation with a modified
  * scale factor and transformed local origin. Can be used to convert
  * between OSGB National Grid and lat,long coordinate systems. See
  * <a href="http://www.gps.gov.uk/guidea.asp" target="_new">
  * www.gps.gov.uk/guidea.asp</a> for details of the transformation.
  * @author Jo Wood.
  * @version 2.2, 8th July, 2004.
  */
//  ***********************************************************

public class OSGB extends UTM
{
    // --------------------- Object variables ---------------------
    
    private double scaleFactor,         // Prime meridian scale factor.
                   latOrigin,lngOrigin,
                   xOffset, yOffset;
    private int direction;
    
    private Ellipsoid wgs84,airy1830;
            
            
    /** Used for testing accuracy of transformation.
      * @param args Command line arguments (ignored).
      */      
    public static void main(String args[])
    {
        OSGB trans = new OSGB(FROM_LAT_LONG);
        Footprint fp = new Footprint(1.717921583f,52.65757031f);
        System.out.println("Original is: "+fp);
        System.out.println("Trans is: "+trans.transformCoords(fp));
        
        Footprint fp2 = new Footprint(651409.903f,313177.270f);
        System.out.println("Inv Trans is: "+trans.invTransformCoords(fp2));
    }
                                
    // ------------------------ Constructors -----------------------
    
    /** Initialises the transformer assuming a forward transformation is into OSGB
      * coordinates. This is equivalent to calling the constructor with FROM_LAT_LONG
      * as the parameter. Assumes that geographic data use the WGS84 ellipsoid. The 
      * transformer will convert data to/from the AIRY_1830 ellipsoid and perform a
      * UTM transformation with a scale factor of 0.9996012717, geographic origin of
      * 49N, 2W and local OSGB origin of E = 400 000, N = -100 000.
      */
    public OSGB()
    {
        this(FROM_LAT_LONG);
    }
        
    /** Initialises the transformer. Assumes that geographic data use
      * the WGS84 ellipsoid. The transformer will convert data to/from the
      * AIRY_1830 ellipsoid and perform a UTM transformation with a
      * scale factor of 0.9996012717, geographic origin of 49N, 2W and 
      * local OSGB origin of E = 400 000, N = -100 000.
      * @param direction Indicates whether tranforming to or from lat/long.
      */
    public OSGB(int direction)
    {
        super(new Ellipsoid(Ellipsoid.AIRY_1830),50,-2);
        wgs84 = new Ellipsoid(Ellipsoid.WGS_84);
        airy1830 = getEllipsoid();
        this.direction = direction;
        scaleFactor = 0.9996012717;
        latOrigin   =  49;
        lngOrigin   =  -2;
        xOffset     =  400000;
        yOffset     = -100000;
    }
    
    // ------------------------- Methods --------------------------
    
    /** Performs a forward lat/long to OSGB transform on the given location.
      * @param fpIn Incoming footprint to transform.
      * @return Transformed footprint.
      */     
    public Footprint transformCoords(Footprint fpIn)
    {
        if (direction == FROM_LAT_LONG)
        {
            // Convert from WGS84 to Airy 1830
            Footprint fp = wgs84.projectDatum(fpIn,Ellipsoid.AIRY_1830);
            
            // Project from geographic to OSBG coordinates.
            return latLongToUTM(fp.getXOrigin(),fp.getYOrigin(),
                                lngOrigin,latOrigin,
                                xOffset,yOffset,
                                scaleFactor);
        }
        else
        {
            // Project from OSGB to geographic coordinates.
            Footprint fp = UTMToLatLong(fpIn.getXOrigin(),fpIn.getYOrigin(),
                                        lngOrigin,latOrigin,
                                        xOffset,yOffset,
                                        scaleFactor);
                                
            // Convert from Airy 1830 to WGS84
            return airy1830.projectDatum(fp,Ellipsoid.WGS_84);
        }
    }
    
    /** Performs an inverse OSGB to lat/long transform on the given location.
      * @param fpIn Incoming footprint to transform.
      * @return Transformed footprint.
      */     
    public Footprint invTransformCoords(Footprint fpIn)
    {
        if (direction == FROM_LAT_LONG)
        {
            // Project from OSGB to geographic coordinates.
            Footprint fp = UTMToLatLong(fpIn.getXOrigin(),fpIn.getYOrigin(),
                                        lngOrigin,latOrigin,
                                        xOffset,yOffset,
                                        scaleFactor);
                                
            // Convert from Airy 1830 to WGS84
            return airy1830.projectDatum(fp,Ellipsoid.WGS_84);
        }
        else
        {
            // Convert from WGS84 to Airy 1830
            Footprint fp = wgs84.projectDatum(fpIn,Ellipsoid.AIRY_1830);
            
            // Project from geographic to OSBG coordinates.
            return latLongToUTM(fp.getXOrigin(),fp.getYOrigin(),
                                lngOrigin,latOrigin,
                                xOffset,yOffset,
                                scaleFactor);
        }
    }
    
    /** Provides a general description of the transformation.
      * @return Description of the transformation.
      */
    public String getDescription()
    {
        if (direction == FROM_LAT_LONG)
            return new String("Lat/long to Ordnance Survey GB National Grid transformation.");
        else
            return new String("Ordnance Survey GB National Grid to lat/long transformation.");
    }
}