package jwo.landserf.process.proj;

import jwo.landserf.structure.*;         // For coordinate footprints.
import jwo.landserf.process.*;

//  ***********************************************************
/** Class for representing Oblique Mercator projections.
  * Includes forward and inverse transforms from/to lat,long. 
  * Uses transformation equations from Snyder (1987) Atlas of Map 
  * Projections summarised at this
  * <a href="http://www.remotesensing.org/geotiff/proj_list/hotine_oblique_mercator.html target="new">
  * Hotine Oblique Mercator</a> page. 
  * @author Jo Wood.
  * @version 2.2, 27th July, 2004.
  */
//  ***********************************************************

public class ObliqueMercator implements Transformable
{
    // ---------------- Object and class variables ----------------
    
    private Ellipsoid ellipsoid;  // Reference ellipsoid.
  
    private boolean doInterpolation;
    private int direction;        // Direction of forward transformation.
    
    private static final double RAD2DEG = 180.0/Math.PI;
    private static final double DEG2RAD = Math.PI/180.0;
    private static final double PI_OVER_2 = Math.PI/2.0;
    private static final double PI_OVER_4 = Math.PI/4.0;
    
                                 /** Indicates a transformation from lat/lng */
    public static final int FROM_LAT_LONG = 1;
    
                                 /** Indicates a transformation to lat/lng */
    public static final int TO_LAT_LONG = 2;
    
    
    // Projection parameters.
    double fc,      // Latitude of centre of the projection.
           lc,      // Longitude of centre of the projection.
           ac,      // Azimuth (true) of the centre line passing through the centre of the projection.
           gc,      // Rectified bearing of the centre line.
           singc,cosgc,
           kc,      // Scale factor at the centre of the projection.

           //Ec,      // False Easting at the centre of projection.
           //Nc,      // False Northing at the centre of projection.

           FE,      // False Easting at the natural origin.
           FN,      // False Northing at the natural origin.
           uc,vc,
           e,e2,    // Ellipsoid eccentricity.
           A,B,H,   // Projection intermediate constants.
           l0,g0,sing0,cosg0;
           
    // ----------------------- Starter Method ---------------------
    
    public static void main(String args[])
    {
        ObliqueMercator om = new ObliqueMercator(new Ellipsoid(Ellipsoid.BESSEL_1841));
        Footprint fp = new Footprint(9.56f,46.924965f);
        //Footprint fp = new Footprint((float)(1.997871312*RAD2DEG),(float)(0.081258569*RAD2DEG));
        Footprint fp2 = om.transformCoords(fp);
        Footprint fp3 = om.invTransformCoords(fp2);
        System.out.println("Start: "+fp);
        System.out.println("om   : "+fp2);
        System.out.println("Back : "+fp3);
        
        System.out.println("Swiss inv: "+om.invTransformCoords(new Footprint(751000,244000)));
        
    }

    // ------------------------ Constructor -----------------------

    /** Initialises the Oblique Mercator converter with the given ellipsoid and
      * projection parameters. The forward transformation will be from Lat/long 
      * to Oblique Merctor.  
      * @param ellipsoid to use in projection.
      */
    public ObliqueMercator(Ellipsoid ellipsoid)
    {
        this.ellipsoid = ellipsoid;
        doInterpolation = true; 
        direction = FROM_LAT_LONG; 
        
        
        // Swiss Ch-03 parameters.
        
        this.ellipsoid = new Ellipsoid(Ellipsoid.BESSEL_1841);
        lc = 7.4395833333333333*DEG2RAD;  // 7 degrees, 26 minutes, 22.5 seconds.
        fc = 46.952405555555556*DEG2RAD;  // 46 degrees, 57 minutes, 8.66 seconds. 
        ac = 90*DEG2RAD;
        gc = 90*DEG2RAD;                          // Assume rectified bearing same as centre line bearing.
        kc = 1;
        FE = -7419820.5907;
        FN = 1200000.0000;
        //FE = 0;
        //FN = 0;  
        
        
        // Test parameters.
        /*
        ellipsoid = new Ellipsoid(Ellipsoid.EVEREST);
        this.ellipsoid = ellipsoid;
        lc = 2.007128640;
        fc = 0.069813170;
        ac = 0.930536611;
        gc = 0.927295218;
        kc = 0.99984;
        //Ec = 0;
        //Nc = 0;
        FE = 400000;
        FN = 100000;
        */
        
        // Perform initial constant calculations
        double cosfc = Math.cos(fc);
        double sinfc = Math.sin(fc);
        singc = Math.sin(gc);
        cosgc = Math.cos(gc);
        e2 = ellipsoid.getSquaredEccentricity();
        e = Math.sqrt(e2);
        double a  = ellipsoid.getEquatorialRadius();
        
        B  = Math.sqrt(1.0 + e2*cosfc*cosfc*cosfc*cosfc / (1-e2));
        A  = a*B*kc*Math.sqrt(1-e2) / (1-e2*sinfc*sinfc);
        double t0 = Math.tan(PI_OVER_4 - fc/2) / Math.pow((1-e*sinfc) / (1+e*sinfc),e/2);
        double D  = B*Math.sqrt(1-e2) / (cosfc*Math.sqrt(1-e2*sinfc*sinfc)); 
        double DSq = D*D;
        if (DSq <1)
            DSq = 1;     // To avoid problems with computation of F below.
        double F  = D + Math.sqrt(DSq -1);
        if (fc <0)      // For centre south of equator.
            F *= -1;
        H = F*Math.pow(t0,B); 
        double G = (F - 1/F)/2;
        sing0 = Math.sin(ac)/D;
        g0 = Math.asin(sing0);
        cosg0 = Math.cos(g0);
        
        if (ac == 90*DEG2RAD)
        {
            l0 = lc - (PI_OVER_2/B);
            System.out.println("special case: "+l0); 
        } 
        else
        {
            l0 = lc -(Math.asin(G*Math.tan(g0)))/B;
            System.out.println("Not special case: "+l0); 
        } 

        vc = 0;
        
        if (ac == 90*DEG2RAD)
            uc = A *(lc - l0);
        else  
        {
            uc = (A/B)*Math.atan(Math.sqrt(DSq-1) / Math.cos(ac));
            if (fc < 0)     // For centre south of equator.
                uc *=-1;
        } 
        
        System.out.println("\tuc: "+uc);
    }
    
    // ------------------------- Methods --------------------------
    
    /** Performs a forward lat/long to Oblique Mercator transform on the given location.
      * @param fpIn Incoming footprint to transform.
      * @return Transformed footprint.
      */     
    public Footprint transformCoords(Footprint fpIn)
    {
        if (direction == FROM_LAT_LONG)
            return latLongToObliqueMercator(fpIn.getXOrigin(),fpIn.getYOrigin());
        else
            return obliqueMercatorToLatLong(fpIn.getXOrigin(),fpIn.getYOrigin());
    }
    
    /** Performs an inverse Oblique Mercator to lat/long transform on the given location.
      * @param fpIn Incoming footprint to transform.
      * @return Transformed footprint.
      */     
    public Footprint invTransformCoords(Footprint fpIn)
    {
        if (direction == FROM_LAT_LONG)
            return obliqueMercatorToLatLong(fpIn.getXOrigin(),fpIn.getYOrigin());
        else
            return latLongToObliqueMercator(fpIn.getXOrigin(),fpIn.getYOrigin());
    }
    
    /** Provides a general description of the transformation.
      * @return Description of the transformation.
      */
    public String getDescription()
    {
        if (direction== FROM_LAT_LONG)
            return new String("Lat/long to Oblique Mercator transformation.");
        else
            return new String("Oblique Mercator to lat/long transformation.");
    }
    
    /** Indicates whether the transformation should use nearest neighbour (false)
      * or some interpolator (true).
      * @return True if transformation should perform some local interpolation.
      */
    public boolean doInterpolation()
    {
        return doInterpolation; 
    }
    
    /** Sets whether the transformation should use nearest neighbour (false)
      * or some interpolator (true).
      * @param doInterpolation True if transformation should perform some local interpolation.
      */
    public void setInterpolation(boolean doInterpolation)
    {
        this.doInterpolation = doInterpolation; 
    }
       
    /** Converts the given (lat,lng) coordinates into an oblique Mercator location.
      * Lat/lng should be in decimal degrees with +ve North and +ve East. 
      * @param lng Longitude (+-180)
      * @param lat Latitude (+-90)
      * @return Oblique Merctator location.
      */ 
    public Footprint latLongToObliqueMercator(float lng, float lat)
    {
        double lngRad = lng*DEG2RAD;
        double latRad = lat*DEG2RAD;
        
        double t = Math.tan(PI_OVER_4 - latRad/2) / Math.pow((1 - e*Math.sin(latRad)) / (1 + e*Math.sin(latRad)),e/2);
        double Q = H / Math.pow(t,B);
        double S = (Q - 1/Q) / 2;
        double T = (Q + 1/Q) / 2;
        double V = Math.sin(B*(lngRad - l0));
        double U = (-V* Math.cos(g0) + S*sing0) / T;
        double v = A*Math.log((1-U)/(1+U))/ (2*B);
        //double u = A*Math.atan((S*Math.cos(g0) + V*sing0) / Math.cos(B*(lngRad-l0)))/B;
        
        double u = A*Math.atan2((S*cosg0 + V*sing0), Math.cos(B*(lngRad-l0)))/B;
        
        //u = u-uc;

        double E = v*cosgc + u*singc + FE;  //Ec;
        double N = u*cosgc - v*singc + FN;  //Nc;
        
        System.out.println("\tt:"+t+" Q:"+Q+" S:"+S+" T:"+T+" V:"+V);
        
        System.out.println("\tv:"+v+" u:"+u);

        return new Footprint((float)E,(float)N);  
    }
    
    
    /** Converts the given Oblique Mercator coordinates into a a lat/long location.
      * Lat/lng will be in decimal degrees with +ve North and +ve East. 
      * @param easting Easting
      * @param northing Northing
      * @return Lat/long location.
      */ 
    public Footprint obliqueMercatorToLatLong(float easting, float northing)
    {
        double vPrime = (easting-FE)*cosgc - (northing-FN)*singc;
        double uPrime = (northing-FN)*cosgc + (easting-FE)*singc;

        double QPrime = Math.pow(Math.E,-B*vPrime/A);
        double SPrime = (QPrime - 1/QPrime) / 2;
        double TPrime = (QPrime + 1/QPrime) / 2;
        double VPrime = Math.sin(B*uPrime/A);
        double UPrime = (VPrime*cosg0 + SPrime*sing0)/TPrime;
        double tPrime = Math.pow(H / Math.sqrt((1 + UPrime) / (1-UPrime)),1/B);
        double c      = PI_OVER_2 - 2*Math.atan(tPrime);
        double lat    = c + Math.sin(2*c)*(e2/2 + 5*e2*e2/24 + e2*e2*e2/12 + 13*e2*e2*e2*e2/360) +
                            Math.sin(4*c)*(7*e2*e2/48 + 29*e2*e2*e2/240 + 811*e2*e2*e2*e2/11520) +
                            Math.sin(6*c)*(7*e2*e2*e2/120 + 81*e2*e2*e2*e2/1120) + 
                            Math.sin(8*c)*(4279*e2*e2*e2*e2/161280);
        double lng    = l0 - Math.atan((SPrime*cosg0 - VPrime*sing0) / Math.cos(B*uPrime/A))/B;

        return new Footprint((float)(lng*RAD2DEG),(float)(lat*RAD2DEG));  
    }
     
  
    /** Reports the ellipsoid used in the transformation.
      * @return Ellipsoid used in the Oblique Mercator transformation.
      */
    public Ellipsoid getEllipsoid()
    {
        return ellipsoid; 
    }
}
