package jwo.landserf.process.proj;

import java.io.*;                      // To make class serializable.

//  **************************************************************
/** Stores information about a map projection. Includes projection
  * type, zone (where relevant) and ellipsoid used.
  * @author Jo Wood
  * @version 2.2, 25th August, 2004.
  */
//  **************************************************************

public class Projection implements Serializable
{
    // --------------------- Object and class Variables ----------------------
    
    private Ellipsoid ellipsoid;      // Ellipsoid used in projection.
    private int type;                 // Type of projection.
    private int zoneNumber;           // UTM zone number (longitudinal).
    private char zoneLetter;          // UTM zone letter (latitudinal).

    // Note: Do not change the values of these constants as they will be 
    //       serialized with the spatial object.
    
                    /** Indicates map projection is undefined. */
    public static final int UNDEFINED = 0;
                    /** Indicates lat/long coordinate system. */
    public static final int LAT_LONG = 1;
                    /** Indicates UTM map projection. */
    public static final int UTM = 2;
                    /** Indicates Ordnance Survey National Grid map projection. */
    public static final int OSGB = 3;
                    /** Indicates State Plane map projection. */
    public static final int STATE_PLANE = 4;
                    /** Indicates Swiss grid map projection. */
    public static final int SWISS = 5;
    
                    /** Indicates Mayhew travel time-space projection. */
    public static final int TIME_SPACE = 101;
    
                    /** Used to ensure consistency when serializing and deserializing. */        
    static final long serialVersionUID = 3353942177225041352L;
 
    // ----------------------- Constructors -------------------------
    
    /** Creates an undefined projection.
      */
    public Projection()
    {
        this(UNDEFINED,null);
    }
    
    /** Creates a projection of the given type with no ellipsoid defined.
      * @param type Type of projection. 
      */
    public Projection(int type)
    {
        this(type,null);
    }
    
    /** Creates a copy of the given projection.
      * @param oldProjection Projection information to copy. 
      */
    public Projection(Projection oldProjection)
    {
        this (oldProjection==null? UNDEFINED:oldProjection.getType(),
              oldProjection==null? null:oldProjection.getEllipsoid());
        
        if (oldProjection != null)      
            setZone(oldProjection.getZoneNumber(),oldProjection.getZoneLetter());
    }
    
    /** Stores the given projection information.
      * @param type Type of projection. 
      * @param ellipsoid Type of ellipsoid used in projection.
      */
    public Projection(int type, Ellipsoid ellipsoid)
    {
        this.type = type;
        this.ellipsoid = ellipsoid;
        zoneNumber = 0;
        zoneLetter = ' ';
    }
 
    // ------------------------- Methods ---------------------------
    
    /** Reports the type of projection stored.
      * @return Projection type. 
      */
    public int getType()
    {
        return type;
    }

    /** Reports the type of ellipsoid stored.
      * @return Ellipsoid used for global map projection or null if not defined. 
      */
    public Ellipsoid getEllipsoid()
    {
        return ellipsoid;
    }
    
    /** Sets the zone number and letter for a UTM projection.
      * @param zoneNumber Zone number to use.
      * @param zoneLetter Zone letter to use.
      */ 
    public void setZone(int zoneNumber, char zoneLetter)
    {
        this.zoneNumber = zoneNumber;
        this.zoneLetter = zoneLetter;
    }
    
    /** Sets the zone number for a UTM projection. 
      * @param zoneNumber Zone number to use.
      */ 
    public void setZoneNumber(int zoneNumber)
    {
        setZone(zoneNumber,' ');
    }
    
    /** Reports the UTM zone number of the given (lat,lng) location. Numbers vary
      * from 1 to 60, where 1 is around 175W, and 60 is 175E. Returns 0 if not defined.
      * @return UTM zone number (longitudinal zone).
      */ 
    public int getZoneNumber()
    {
        return zoneNumber;
    }
    
    /** Reports the UTM zone letter of the given (lat,lng) location. Letters vary 
      * from N,M,O...X going north from equator and M,L,K...C going south from
      * equator. Returns Z if outside 84N or 80S or ' ' (space) if not defined. 
      * @return UTM zone letter (latitudinal zone).
      */ 
    public char getZoneLetter()
    {
        return zoneLetter;
    }

    /** Reports the full UTM zone of the given (lat,lng) location.
      * @return UTM zone (e.g 19F for southern tip of South America).
      */ 
    public String getZone()
    {
        return new String(Integer.toString(zoneNumber))+Character.toString(zoneLetter);
    }
    
    /** Returns a human-readable description of this projection.
      * @return Description of this projection.
      */ 
    public String toString()
    {
        String typeName = null;
        
        if (type == UTM)
            typeName = new String(getName(type)+ " zone "+getZone());
        else
            typeName = getName(type);
        
        if (ellipsoid == null)
            return (typeName);
        else
            return (typeName+" ("+ellipsoid.getName()+" ellipsoid)");    
    }
    
    /** Reports the name of the given projection type.
      * @param projType Type of projection to name. 
      * @return Name of given projection type. Useful for reporting to a GUI.
      */ 
    public static String getName(int projType)
    {
        String typeName = null;

        switch (projType)
        {
            case LAT_LONG:
                typeName = new String("Latitude/longitude");
                break;
            case UTM:
                typeName = new String("UTM");
                break;
            case OSGB:
                typeName = new String("Ordnance Survey National Grid");
                break;
            case STATE_PLANE:
                typeName = new String("US State Plane");
                break;
            case SWISS:
                typeName = new String("Swiss National Grid");
                break;
            case TIME_SPACE:
                typeName = new String("Mayhew time-space");
                break;
            default:
                typeName = new String("Undefined projection");
        }
        return typeName;
    }
}
    
    